"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$equal = exports.$max = exports.$min = exports.$compare = exports.$datecompare = exports.$numcompare = void 0;
const commons_1 = require("./commons");
const tsdate_1 = require("./tsdate");
const types_1 = require("./types");
function $numcompare(a, b) {
    if (isNaN(a) || isNaN(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    return a < b ? types_1.Ascending : types_1.Descending;
}
exports.$numcompare = $numcompare;
function $datecompare(a, b) {
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    /* TODO: WE SHOULD BE PERMITED NOT TO CAST a or b in new TSDate() */
    if (!(a instanceof tsdate_1.TSDate)) {
        a = new tsdate_1.TSDate(a /* this is wrong  */);
    }
    if (!(b instanceof tsdate_1.TSDate)) {
        b = new tsdate_1.TSDate(b /* this is wrong  */);
    }
    return a.compare(b);
}
exports.$datecompare = $datecompare;
function $compare(a, b) {
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b)) {
        return undefined;
    }
    if (a === b) {
        return types_1.Same;
    }
    if (typeof a === 'number' && typeof b === 'number') {
        return $numcompare(a, b);
    }
    if ((0, commons_1.$isstring)(a) && (0, commons_1.$isstring)(b)) {
        // before, we used Buffer.compare() to make bytes comparison
        return a > b ? types_1.Descending : (a < b ? types_1.Ascending : types_1.Same);
    }
    if ((0, commons_1.$isarray)(a) && (0, commons_1.$isarray)(b)) {
        const na = a.length, nb = b.length;
        let i = 0;
        while (i < na && i < nb) {
            const c = $compare(a[i], b[i]);
            if (c !== types_1.Same) {
                return c;
            }
            i++;
        }
        return na === nb ? types_1.Same : (i < na ? types_1.Descending : types_1.Ascending);
    }
    if ((a instanceof Date || a instanceof tsdate_1.TSDate) && (b instanceof Date || b instanceof tsdate_1.TSDate)) {
        return $datecompare(a, b);
    }
    if (a instanceof Buffer && b instanceof Buffer) {
        return Buffer.compare(a, b);
    }
    if ((a instanceof ArrayBuffer || ArrayBuffer.isView(a)) && (b instanceof ArrayBuffer || ArrayBuffer.isView(b))) {
        a = new Uint8Array(a);
        b = new Uint8Array(b);
        const na = a.length, nb = b.length;
        let i = 0;
        while (i < na && i < nb) {
            const c = $numcompare(a[i], b[i]);
            if (c !== types_1.Same) {
                return c;
            }
            i++;
        }
        return na === nb ? types_1.Same : (i < na ? types_1.Descending : types_1.Ascending);
    }
    return (0, commons_1.$isobject)(a) && ('compare' in a) ? a.compare(b) : undefined;
}
exports.$compare = $compare;
function $min(values) { return _minmax(values, types_1.Descending); }
exports.$min = $min;
function $max(values) { return _minmax(values, types_1.Ascending); }
exports.$max = $max;
function _minmax(values, compValue) {
    let ret = undefined;
    if ((0, commons_1.$ok)(values)) {
        for (let v of values) {
            if (!(0, commons_1.$ok)(v)) {
                return undefined;
            }
            else if (!(0, commons_1.$ok)(ret)) {
                ret = v;
            }
            else {
                let comp = $compare(ret, v);
                if (!(0, commons_1.$defined)(comp)) {
                    return undefined;
                }
                else if (comp === compValue) {
                    ret = v;
                }
            }
        }
    }
    return ret;
}
function $equal(a, b) {
    if (a === b) {
        return true;
    }
    if (typeof a === 'number' && typeof b === 'number') {
        return a === b;
    } // in order to cover NaN inequality and infinity equality
    if (!(0, commons_1.$ok)(a) || !(0, commons_1.$ok)(b))
        return false;
    if ((0, commons_1.$isarray)(a) && (0, commons_1.$isarray)(b)) {
        const n = a.length;
        if (n !== b.length)
            return false;
        for (let i = 0; i < n; i++) {
            if (!$equal(a[i], b[i]))
                return false;
        }
        return true;
    }
    if (a instanceof Date && b instanceof Date) {
        return a.getTime() === b.getTime();
    }
    if ((a instanceof Date || a instanceof tsdate_1.TSDate) && (b instanceof Date || b instanceof tsdate_1.TSDate)) {
        if (a instanceof Date) {
            a = new tsdate_1.TSDate(a);
        }
        if (b instanceof Date) {
            b = new tsdate_1.TSDate(b);
        }
        return a.isEqual(b);
    }
    if ((0, commons_1.$isobject)(a) && ('isEqual' in a)) {
        return a.isEqual(b);
    }
    if ((0, commons_1.$isobject)(b) && ('isEqual' in b)) {
        return b.isEqual(a);
    }
    if (a instanceof Set && b instanceof Set) {
        return a.size === b.size && [...a.keys()].every(e => b.has(e));
    }
    if (a instanceof Map && b instanceof Map) {
        const ak = a.keys();
        const bk = b.keys();
        const keys = a.size >= b.size ? ak : bk;
        // we may have different expressed keys with undefined as value...
        // eg: MapA{a:1, b:undefined} equals MapB{a:1} since MapB.get('b') returns undefined 
        for (let k of keys) {
            if (!$equal(a.get(k), b.get(k)))
                return false;
        }
        return true;
    }
    if (a instanceof Buffer && b instanceof Buffer) {
        return Buffer.compare(a, b) === 0;
    }
    if ((a instanceof ArrayBuffer || ArrayBuffer.isView(a)) && (b instanceof ArrayBuffer || ArrayBuffer.isView(b))) {
        a = new Uint8Array(a);
        b = new Uint8Array(b);
        const n = a.length;
        if (n !== b.length)
            return false;
        for (let i = 0; i < n; i++) {
            if (a[i] !== b[i])
                return false;
        }
        return true;
    }
    if (Object.getPrototypeOf(a) === Object.prototype && Object.getPrototypeOf(b) === Object.prototype) {
        const ak = Object.getOwnPropertyNames(a);
        const bk = Object.getOwnPropertyNames(a);
        const keys = ak.length >= bk.length ? ak : bk;
        // we may have different expressed keys with undefined as value...
        // eg: {a:1, b:undefined} equals {a:1}
        for (let k of keys) {
            if (!$equal(a[k], b[k]))
                return false;
        }
        return true;
    }
    return false;
}
exports.$equal = $equal;
if (!('min' in Array.prototype)) {
    Array.prototype.min = function min() { return $min(this); };
}
if (!('max' in Array.prototype)) {
    Array.prototype.max = function max() { return $max(this); };
}
//# sourceMappingURL=compare.js.map