import { TSClone, TSObject } from "./tsobject";
import { Comparison, uint, uint8 } from "./types";
/**
 * TSColor are immutable objects. RGB Colors may be cached,
 * CMYK and Grayscale colors are not. It's a one class implementation
 * of three colorspace. It's agains all OOD rules but it's efficient
 * in that case because of the proximity between CMYK and Grayscale
 * and the usage of primitives methods like rgb() or grayComponent() or
 * cmykComponents()
 */
export interface TSColorToStringOptions {
    removeAlpha?: true;
    shortestCSS?: true;
    rgbaCSSLike?: true;
    colorSpace?: TSColorSpace;
}
/**
 *  Those are the 3 colorspace we can use.
 *  If you need HSL or HSB values for a color,
 *  use the hsb() or hsl() instance methods to
 *  get the values you need.
 */
export declare enum TSColorSpace {
    RGB = "RGB",
    CMYK = "CMYK",
    Grayscale = "GRAYSCALE"
}
export declare enum TSToGrayScaleMode {
    Brightess = 0,
    Luminosity = 1,
    Luminance = 2
}
export declare class TSColor implements TSObject, TSClone<TSColor> {
    readonly colorSpace: TSColorSpace;
    private _channels;
    private _alpha;
    private _name?;
    private static __colorsCache;
    private static readonly __colorCacheMaxSize;
    static readonly black: TSColor;
    static readonly white: TSColor;
    static readonly cyan: TSColor;
    static readonly magenta: TSColor;
    static readonly yellow: TSColor;
    static readonly red: TSColor;
    static readonly green: TSColor;
    static readonly blue: TSColor;
    static rgb(stringColor: string): TSColor;
    static rgb(colorDefinition: number): TSColor;
    static rgb(R: number, G: number, B: number, A?: number): TSColor;
    static rgbcomponents(r: number, g: number, b: number, opacity?: number): TSColor;
    static cmyk(C: number, M: number, Y: number, K: number, opacity?: number): TSColor;
    static grayscale(whiteIntensity: number, opacity?: number): TSColor;
    private constructor();
    rgb(): [uint8, uint8, uint8];
    rgbComponents(): [number, number, number];
    cmykComponents(): [number, number, number, number];
    grayComponent(): number;
    get name(): string;
    clone(): TSColor;
    get red(): uint8;
    get green(): uint8;
    get blue(): uint8;
    get alpha(): uint8;
    get transparency(): uint8;
    get cyan(): number;
    get magenta(): number;
    get yellow(): number;
    get black(): number;
    get opacity(): number;
    get gray(): number;
    hsl(): [number, number, number];
    hsb(): [number, number, number];
    luminosity(): number;
    luminance(): number;
    isPale(): boolean;
    toAlpha(newAlpha: uint8): TSColor;
    toOpacity(newOpacity: number): TSColor;
    toRGB(): TSColor;
    toCMYK(): TSColor;
    toGrayscale(mode?: TSToGrayScaleMode): TSColor;
    lighterColor(): TSColor;
    darkerColor(): TSColor;
    lightestColor(): TSColor;
    darkestColor(): TSColor;
    matchingColor(): TSColor;
    toNumber(): number;
    toUnsigned(): uint;
    isSimilar(other: TSColor): boolean;
    isEqual(other: any): boolean;
    compare(other: any): Comparison;
    toString(opts?: TSColorToStringOptions): string;
    toJSON(): any;
    toArray(): number[];
    private _modifiedColorsColor;
    private static readonly __TSWebColorNames;
    private static readonly __WebColorsHexParsers;
    private static _cacheRGBColor;
    private static _cachedRGBColor;
    private static _parseHexColorString;
}
export interface TSColorConstructor {
    new (colorSpace: TSColorSpace, channels: number[], alpha: number, name?: string): TSColor;
}
