"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSColor = exports.TSToGrayScaleMode = exports.TSColorSpace = void 0;
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const types_1 = require("./types");
/**
 *  Those are the 3 colorspace we can use.
 *  If you need HSL or HSB values for a color,
 *  use the hsb() or hsl() instance methods to
 *  get the values you need.
 */
var TSColorSpace;
(function (TSColorSpace) {
    TSColorSpace["RGB"] = "RGB";
    TSColorSpace["CMYK"] = "CMYK";
    TSColorSpace["Grayscale"] = "GRAYSCALE";
})(TSColorSpace = exports.TSColorSpace || (exports.TSColorSpace = {}));
var TSToGrayScaleMode;
(function (TSToGrayScaleMode) {
    TSToGrayScaleMode[TSToGrayScaleMode["Brightess"] = 0] = "Brightess";
    TSToGrayScaleMode[TSToGrayScaleMode["Luminosity"] = 1] = "Luminosity";
    TSToGrayScaleMode[TSToGrayScaleMode["Luminance"] = 2] = "Luminance";
})(TSToGrayScaleMode = exports.TSToGrayScaleMode || (exports.TSToGrayScaleMode = {}));
class TSColor {
    /*
        TSColor constructor is now private.
        Use rgb() grayscale() and cmyk() static methods
        to create new colors.
    */
    constructor(colorSpace, channels, alpha, name) {
        this.colorSpace = colorSpace;
        this._channels = channels;
        this._alpha = alpha;
        if ((0, commons_1.$length)(name)) {
            this._name = name;
        }
    }
    static rgb() {
        if (arguments.length === 1) {
            if ((0, commons_1.$isstring)(arguments[0])) {
                const s = (0, commons_1.$ftrim)(arguments[0]).toLowerCase();
                let color = TSColor._cachedRGBColor(s);
                if ((0, commons_1.$defined)(color)) {
                    return color;
                }
                const [channels, alpha] = TSColor._parseHexColorString(s);
                if ((0, commons_1.$ok)(channels)) {
                    color = new TSColor(TSColorSpace.RGB, channels, alpha);
                    TSColor._cacheRGBColor(color);
                }
                return color;
            }
            else if ((0, commons_1.$isunsigned)(arguments[0], types_1.UINT32_MAX)) {
                const v = arguments[0];
                // by calling the static methods, the color will be cached.
                return TSColor.rgb((v >> 16) & 0xff, (v >> 8) & 0xff, v & 0xff, 0xff - ((v >> 24) & 0xff));
            }
        }
        else if (arguments.length === 3 || arguments.length === 4) {
            const R = arguments[0];
            const G = arguments[1];
            const B = arguments[2];
            const A = arguments.length === 4 ? arguments[3] : 0xFF;
            if ((0, commons_1.$isunsigned)(R, 0xff) && (0, commons_1.$isunsigned)(G, 0xff) && (0, commons_1.$isunsigned)(B, 0xff) && (0, commons_1.$isunsigned)(A, 0xff)) {
                const s = _colorToStandardCSS(R, G, B, A);
                let color = TSColor._cachedRGBColor(s);
                if ((0, commons_1.$defined)(color)) {
                    return color;
                }
                color = new TSColor(TSColorSpace.RGB, [R, G, B], A);
                TSColor._cacheRGBColor(color);
                return color;
            }
        }
        throw 'TSColor.rgb() : Bad parameters';
    }
    static rgbcomponents(r, g, b, opacity) {
        opacity = (0, commons_1.$ok)(opacity) ? opacity : 1;
        if ((0, commons_1.$isnumber)(r) && (0, commons_1.$isnumber)(g) && (0, commons_1.$isnumber)(b) && (0, commons_1.$isnumber)(opacity)) {
            r = _component(r);
            g = _component(g);
            b = _component(b);
            opacity = _component(opacity);
            return TSColor.rgb((r * 255) | 0, (g * 255) | 0, (b * 255) | 0, (opacity * 255) | 0);
        }
        throw 'TSColor.rgbcomponents() : Bad parameters';
    }
    static cmyk(C, M, Y, K, opacity) {
        opacity = (0, commons_1.$ok)(opacity) ? opacity : 1;
        if ((0, commons_1.$isnumber)(C) && (0, commons_1.$isnumber)(M) && (0, commons_1.$isnumber)(Y) && (0, commons_1.$isnumber)(K) && (0, commons_1.$isnumber)(opacity)) {
            C = _component(C);
            M = _component(M);
            Y = _component(Y);
            K = _component(K);
            opacity = _component(opacity);
            return new TSColor(TSColorSpace.CMYK, [C, M, Y, K], opacity);
        }
        throw 'TSColor.cmyk() : Bad parameters';
    }
    static grayscale(whiteIntensity, opacity) {
        opacity = (0, commons_1.$ok)(opacity) ? opacity : 1;
        if ((0, commons_1.$isnumber)(whiteIntensity) && (0, commons_1.$isnumber)(opacity)) {
            whiteIntensity = _component(whiteIntensity);
            return new TSColor(TSColorSpace.Grayscale, [0, 0, 0, 1 - whiteIntensity], opacity);
        }
        throw 'TSColor.grayscale() : Bad parameters';
    }
    // ================== primitive methods =======================
    rgb() {
        if (this.colorSpace === TSColorSpace.RGB) {
            return [(this._channels[0] & 0xFF), (this._channels[1] & 0xFF), (this._channels[2] & 0xFF)];
        }
        const [C, M, Y, K] = this.cmykComponents();
        const light = 255.0 * (1 - K);
        function _rc(x) { return ((light * (1 - x)) & 0xFF); }
        return [_rc(C), _rc(M), _rc(Y)];
    }
    rgbComponents() {
        if (this.colorSpace === TSColorSpace.RGB) {
            const [R, G, B] = this.rgb();
            return [R / 255, G / 255, B / 255];
        }
        const [C, M, Y, K] = this.cmykComponents();
        return [(1 - K) * (1 - C), (1 - K) * (1 - M), (1 - K) * (1 - Y)];
    }
    cmykComponents() {
        if (this.colorSpace !== TSColorSpace.RGB) {
            return [this._channels[0], this._channels[1], this._channels[2], this._channels[3]];
        }
        const [r, g, b] = this.rgbComponents();
        let C = 1 - r;
        let M = 1 - g;
        let Y = 1 - b;
        let K = 1;
        if (C < K) {
            K = C;
        }
        if (M < K) {
            K = M;
        }
        if (Y < K) {
            K = Y;
        }
        return K >= 1 ? [0, 0, 0, 1] : [(C - K) / (1 - K), (M - K) / (1 - K), (Y - K) / (1 - K), K];
    }
    grayComponent() {
        if (this.colorSpace === TSColorSpace.RGB) {
            const [, , L] = this.hsb();
            return L / 100;
        }
        let [C, M, Y, K] = this.cmykComponents();
        if (K >= 1) {
            K = 1;
        }
        else if (C !== 0 || M !== 0 || Y !== 0) {
            const [, , L] = this.hsb();
            return L / 100;
        }
        return 1 - K;
    }
    // ================== accessors ===============================
    get name() { return (0, commons_1.$string)(this._name); }
    clone() { return this; } // no clone on immutable objects
    // RGB color space
    get red() { const [R, ,] = this.rgb(); return R; }
    get green() { const [, G,] = this.rgb(); return G; }
    get blue() { const [, , B] = this.rgb(); return B; }
    get alpha() { return (this.colorSpace === TSColorSpace.RGB ? this._alpha : (0, commons_1.$tounsigned)(_component(this._alpha) * 255)); }
    get transparency() { return 255 - this.alpha; }
    // CYMK color space
    get cyan() { const [C, , ,] = this.cmykComponents(); return C; }
    get magenta() { const [, M, ,] = this.cmykComponents(); return M; }
    get yellow() { const [, , Y,] = this.cmykComponents(); return Y; }
    get black() { const [, , , K] = this.cmykComponents(); return K; }
    get opacity() { return this.colorSpace !== TSColorSpace.RGB ? this._alpha : _component(this._alpha / 255); }
    // GRAY color space
    get gray() { return this.grayComponent(); }
    // ================== other methods ===========================
    hsl() {
        const [r, g, b] = this.rgbComponents();
        const l = Math.max(r, g, b);
        const s = l - Math.min(r, g, b);
        const h = s ? (l === r ? (g - b) / s : (l === g ? 2 + (b - r) / s : 4 + (r - g) / s)) : 0;
        return [
            Math.round(60 * (h < 0 ? h + 6 : h)),
            Math.round(100 * (s ? (l <= 0.5 ? s / (2 * l - s) : s / (2 - (2 * l - s))) : 0)),
            Math.round((100 * (2 * l - s)) / 2),
        ];
    }
    hsb() {
        const [r, g, b] = this.rgbComponents();
        if (r === g && g === b) {
            return [0, 0, Math.round(r * 100)];
        }
        const v = Math.max(r, g, b);
        const n = v - Math.min(r, g, b);
        const h = n === 0 ? 0 : n && v === r ? (g - b) / n : v === g ? 2 + (b - r) / n : 4 + (r - g) / n;
        return [
            Math.round(60 * (h < 0 ? h + 6 : h)),
            Math.round(v && (n / v) * 100),
            Math.round(v * 100)
        ];
    }
    luminosity() {
        // formula from https://www.johndcook.com/blog/2009/08/24/algorithms-convert-color-grayscale/
        const [r, g, b] = this.rgbComponents();
        return 0.21 * r + 0.72 * g + 0.07 * b;
    }
    luminance() {
        const [r, g, b] = this.rgbComponents();
        return 0.3 * r + 0.59 * g + 0.11 * b;
    }
    isPale() { return this.luminance() > 0.6; }
    toAlpha(newAlpha) {
        if (!(0, commons_1.$isunsigned)(newAlpha, 0xFF)) {
            throw 'Bad new alpha parameter';
        }
        if (this.colorSpace !== TSColorSpace.RGB) {
            return this.toOpacity(newAlpha / 255.0);
        }
        if (newAlpha === this._alpha) {
            return this;
        }
        const [R, G, B] = this.rgb();
        return TSColor.rgb(R, G, B, newAlpha);
    }
    toOpacity(newOpacity) {
        newOpacity = _component(newOpacity);
        if (this.colorSpace === TSColorSpace.RGB) {
            return this.toAlpha(((newOpacity * 255.0) | 0));
        }
        if (newOpacity === this._alpha) {
            return this;
        }
        const [C, M, Y, K] = this.cmykComponents();
        return this.colorSpace === TSColorSpace.Grayscale ? TSColor.grayscale(1 - K, newOpacity) : TSColor.cmyk(C, M, Y, K, newOpacity);
    }
    toRGB() {
        if (this.colorSpace === TSColorSpace.RGB) {
            return this;
        }
        const [R, G, B] = this.rgb();
        return TSColor.rgb(R, G, B, this.alpha);
    }
    toCMYK() {
        if (this.colorSpace === TSColorSpace.CMYK) {
            return this;
        }
        const [C, M, Y, K] = this.cmykComponents();
        return TSColor.cmyk(C, M, Y, K, this.opacity);
    }
    // usesLuminosity = true if you want another more precise RGB or CMYK to grayscale conversion 
    toGrayscale(mode = TSToGrayScaleMode.Brightess) {
        if (this.colorSpace === TSColorSpace.Grayscale) {
            return this;
        }
        return TSColor.grayscale(mode === TSToGrayScaleMode.Brightess ?
            this.grayComponent() :
            (mode === TSToGrayScaleMode.Luminosity ? this.luminosity() : this.luminance()));
    }
    lighterColor() { return this._modifiedColorsColor(1, _cmykLighter, _rgbLighter); }
    darkerColor() { return this._modifiedColorsColor(1, _cmykDarker, _rgbDarker); }
    lightestColor() { return this._modifiedColorsColor(2, _cmykLighter, _rgbLighter); }
    darkestColor() { return this._modifiedColorsColor(2, _cmykDarker, _rgbDarker); }
    matchingColor() { return this.isPale() ? this.darkestColor() : this.lightestColor(); }
    toNumber() {
        const [R, G, B] = this.rgb();
        return ((0xff - this.alpha) << 24) | (R << 16) | (G << 8) | B;
    }
    toUnsigned() { return this.toNumber(); }
    isSimilar(other) {
        if (this === other) {
            return true;
        }
        if (other instanceof TSColor) {
            if (this.colorSpace === other.colorSpace) {
                return this._alpha === other.alpha && (0, compare_1.$equal)(this._channels, other._channels);
            }
            if (this.alpha === other.alpha) {
                const [R1, G1, B1] = this.rgb();
                const [R2, G2, B2] = other.rgb();
                return R1 === R2 && G1 === G2 && B1 === B2;
            }
        }
        return false;
    }
    // ============ TSObject conformance =============== 
    isEqual(other) {
        return this === other || (other instanceof TSColor &&
            this.colorSpace === other.colorSpace &&
            this._alpha === other._alpha &&
            (0, compare_1.$equal)(this._channels, other._channels));
    }
    compare(other) {
        return this === other ? types_1.Same :
            (other instanceof TSColor ? (0, compare_1.$numcompare)(this.luminance(), other.luminance()) : undefined);
    }
    toString(opts = {}) {
        const colorSpace = (0, commons_1.$ok)(opts.colorSpace) ? opts.colorSpace : this.colorSpace;
        let opacity;
        if (colorSpace === TSColorSpace.RGB) {
            const [R, G, B] = this.rgb();
            const alpha = this.alpha;
            if (this.alpha === 255 || opts.removeAlpha) {
                if (opts.shortestCSS && _isShortRGB(R, G, B)) {
                    return _colorToShortCSS(R, G, B);
                }
                return _colorToStandardCSS(R, G, B);
            }
            if (opts.rgbaCSSLike) {
                return _colorToStandardCSS(R, G, B, alpha);
            }
            return `rgba(${R},${G},${B},${this.opacity})`;
        }
        opacity = this.opacity;
        const [C, M, Y, K] = this.cmykComponents();
        if (colorSpace === TSColorSpace.Grayscale) {
            return opacity === 1 || opts.removeAlpha ? `gray(${1 - K})` : `gray(${1 - K}, ${opacity})`;
        }
        return opacity === 1 || opts.removeAlpha ? `cmyk(${C}, ${M}, ${Y}, ${K})` : `cmyka(${C}, ${M}, ${Y}, ${K}, ${opacity})`;
    }
    toJSON() {
        switch (this.colorSpace) {
            case TSColorSpace.CMYK:
                const [C, M, Y, K] = this.cmykComponents();
                return { cyan: C, magenta: M, yellow: Y, black: K, opacity: this._alpha };
            case TSColorSpace.Grayscale:
                const [, , , KS] = this.cmykComponents();
                return { grayscale: 1 - KS, opacity: this._alpha };
            case TSColorSpace.RGB:
                const [R, G, B] = this.rgb();
                return this.alpha === 255 ? _colorToStandardCSS(R, G, B) : { red: R, green: G, blue: B, alpha: this._alpha };
        }
    }
    toArray() { return [...this._channels, this._alpha]; }
    // =============== private methods =========================
    _modifiedColorsColor(times, cmykFunction, rgbFunction) {
        if (this.colorSpace === TSColorSpace.RGB) {
            let [R, G, B] = this.rgb();
            for (let i = 0; i < times; i++) {
                R = rgbFunction(R);
                G = rgbFunction(G), B = rgbFunction(B);
            }
            return TSColor.rgb(R, G, B, this._alpha);
        }
        let [C, M, Y, K] = this.cmykComponents();
        for (let i = 0; i < times; i++) {
            C = cmykFunction(C);
            M = cmykFunction(M);
            Y = cmykFunction(Y);
            K = cmykFunction(K);
        }
        return this.colorSpace === TSColorSpace.Grayscale ?
            TSColor.grayscale(1 - K, this._alpha) :
            TSColor.cmyk(C, M, Y, K, this._alpha);
    }
    static _cacheRGBColor(c) {
        if ((0, commons_1.$defined)(TSColor.__colorsCache) && c.colorSpace === TSColorSpace.RGB && TSColor.__colorsCache.size + 7 < TSColor.__colorCacheMaxSize) {
            const [R, G, B] = c.rgb();
            const s = _colorToStandardCSS(R, G, B, c.alpha);
            if (!TSColor.__colorsCache.has(s) && (!c.name.length || !TSColor.__colorsCache.has(c.name))) {
                if (c.name.length) {
                    TSColor.__colorsCache.set(c.name, c);
                }
                TSColor.__colorsCache.set(s, c);
                TSColor.__colorsCache.set(s.slice(1), c);
                if (c.alpha === 0xff) {
                    TSColor.__colorsCache.set(s.slice(0, 7), c);
                    TSColor.__colorsCache.set(s.slice(1, 7), c);
                    if (_isShortRGB(R, G, B)) {
                        const short = _colorToShortCSS(R, G, B);
                        TSColor.__colorsCache.set(short, c);
                        TSColor.__colorsCache.set(short.slice(1), c);
                    }
                }
            }
        }
    }
    static _cachedRGBColor(c) {
        if (!(0, commons_1.$defined)(TSColor.__colorsCache)) {
            TSColor.__colorsCache = new Map();
            for (let name of (0, commons_1.$keys)(TSColor.__TSWebColorNames)) {
                const [channels, alpha] = TSColor._parseHexColorString(TSColor.__TSWebColorNames[name]);
                if ((0, commons_1.$ok)(channels)) {
                    const c = new TSColor(TSColorSpace.RGB, channels, alpha, name);
                    TSColor._cacheRGBColor(c);
                }
            }
            ;
        }
        return TSColor.__colorsCache.get(c);
    }
    static _parseHexColorString(s) {
        const len = (0, commons_1.$length)(s);
        if (len < 10) {
            let parser = TSColor.__WebColorsHexParsers[len];
            if ((0, commons_1.$ok)(parser)) {
                const m = s.match(parser.rx);
                if ((0, commons_1.$ok)(m)) {
                    let channels = [];
                    for (let i = 0; i < 3; i++) {
                        const v = parseInt(m[i + 1], 16);
                        if (!(0, commons_1.$isunsigned)(v, 0xFF)) {
                            return [null, types_1.UINT8_MIN];
                        }
                        channels[i] = v;
                    }
                    if (parser.short) {
                        channels = channels.map(v => ((v << 4) | v));
                    }
                    let alpha = m.length === 5 ? parseInt(m[4], 16) : types_1.UINT8_MAX;
                    return [channels, alpha];
                }
            }
        }
        return [null, types_1.UINT8_MIN];
    }
}
exports.TSColor = TSColor;
TSColor.__colorCacheMaxSize = 16384;
// THOSE ARE PRIMARY Grayscale or CMYK COLORS
// TO GET RGB Colors use TSColor.rgb(). 
// EG: To get a named color called green, use: TSColor.rgb('green')
TSColor.black = TSColor.grayscale(0);
TSColor.white = TSColor.grayscale(1);
TSColor.cyan = TSColor.cmyk(1, 0, 0, 0);
TSColor.magenta = TSColor.cmyk(0, 1, 0, 0);
TSColor.yellow = TSColor.cmyk(0, 0, 1, 0);
TSColor.red = TSColor.cmyk(0, 1, 1, 0);
TSColor.green = TSColor.cmyk(1, 0, 1, 0);
TSColor.blue = TSColor.cmyk(1, 1, 0, 0);
TSColor.__TSWebColorNames = {
    aliceblue: "#f0f8ff",
    antiquewhite: "#faebd7",
    aqua: "#00ffff",
    aquamarine: "#7fffd4",
    azure: "#f0ffff",
    beige: "#f5f5dc",
    bisque: "#ffe4c4",
    black: "#000000",
    blanchedalmond: "#ffebcd",
    blue: "#0000ff",
    blueviolet: "#8a2be2",
    brown: "#a52a2a",
    burlywood: "#deb887",
    cadetblue: "#5f9ea0",
    chartreuse: "#7fff00",
    chocolate: "#d2691e",
    coral: "#ff7f50",
    cornflowerblue: "#6495ed",
    cornsilk: "#fff8dc",
    crimson: "#dc143c",
    cyan: "#00ffff",
    darkblue: "#00008b",
    darkcyan: "#008b8b",
    darkgoldenrod: "#b8860b",
    darkgray: "#a9a9a9",
    darkgreen: "#006400",
    darkgrey: "#a9a9a9",
    darkkhaki: "#bdb76b",
    darkmagenta: "#8b008b",
    darkolivegreen: "#556b2f",
    darkorange: "#ff8c00",
    darkorchid: "#9932cc",
    darkred: "#8b0000",
    darksalmon: "#e9967a",
    darkseagreen: "#8fbc8f",
    darkslateblue: "#483d8b",
    darkslategray: "#2f4f4f",
    darkslategrey: "#2f4f4f",
    darkturquoise: "#00ced1",
    darkviolet: "#9400d3",
    deeppink: "#ff1493",
    deepskyblue: "#00bfff",
    dimgray: "#696969",
    dimgrey: "#696969",
    dodgerblue: "#1e90ff",
    firebrick: "#b22222",
    floralwhite: "#fffaf0",
    forestgreen: "#228b22",
    fuchsia: "#ff00ff",
    gainsboro: "#dcdcdc",
    ghostwhite: "#f8f8ff",
    goldenrod: "#daa520",
    gold: "#ffd700",
    gray: "#808080",
    green: "#00ff00",
    greenyellow: "#adff2f",
    grey: "#808080",
    honeydew: "#f0fff0",
    hotpink: "#ff69b4",
    indianred: "#cd5c5c",
    indigo: "#4b0082",
    ivory: "#fffff0",
    khaki: "#f0e68c",
    lavenderblush: "#fff0f5",
    lavender: "#e6e6fa",
    lawngreen: "#7cfc00",
    lemonchiffon: "#fffacd",
    lightblue: "#add8e6",
    lightcoral: "#f08080",
    lightcyan: "#e0ffff",
    lightgoldenrodyellow: "#fafad2",
    lightgray: "#d3d3d3",
    lightgreen: "#90ee90",
    lightgrey: "#d3d3d3",
    lightpink: "#ffb6c1",
    lightsalmon: "#ffa07a",
    lightseagreen: "#20b2aa",
    lightskyblue: "#87cefa",
    lightslategray: "#778899",
    lightslategrey: "#778899",
    lightsteelblue: "#b0c4de",
    lightyellow: "#ffffe0",
    lime: "#00ff00",
    limegreen: "#32cd32",
    linen: "#faf0e6",
    magenta: "#ff00ff",
    maroon: "#800000",
    mediumaquamarine: "#66cdaa",
    mediumblue: "#0000cd",
    mediumorchid: "#ba55d3",
    mediumpurple: "#9370db",
    mediumseagreen: "#3cb371",
    mediumslateblue: "#7b68ee",
    mediumspringgreen: "#00fa9a",
    mediumturquoise: "#48d1cc",
    mediumvioletred: "#c71585",
    midnightblue: "#191970",
    mintcream: "#f5fffa",
    mistyrose: "#ffe4e1",
    moccasin: "#ffe4b5",
    navajowhite: "#ffdead",
    navy: "#000080",
    oldlace: "#fdf5e6",
    olive: "#808000",
    olivedrab: "#6b8e23",
    orange: "#ffa500",
    orangered: "#ff4500",
    orchid: "#da70d6",
    palegoldenrod: "#eee8aa",
    palegreen: "#98fb98",
    paleturquoise: "#afeeee",
    palevioletred: "#db7093",
    papayawhip: "#ffefd5",
    peachpuff: "#ffdab9",
    peru: "#cd853f",
    pink: "#ffc0cb",
    plum: "#dda0dd",
    powderblue: "#b0e0e6",
    purple: "#800080",
    rebeccapurple: "#663399",
    red: "#ff0000",
    rosybrown: "#bc8f8f",
    royalblue: "#4169e1",
    saddlebrown: "#8b4513",
    salmon: "#fa8072",
    sandybrown: "#f4a460",
    seagreen: "#2e8b57",
    seashell: "#fff5ee",
    sienna: "#a0522d",
    silver: "#c0c0c0",
    skyblue: "#87ceeb",
    slateblue: "#6a5acd",
    slategray: "#708090",
    slategrey: "#708090",
    snow: "#fffafa",
    springgreen: "#00ff7f",
    steelblue: "#4682b4",
    tan: "#d2b48c",
    teal: "#008080",
    thistle: "#d8bfd8",
    tomato: "#ff6347",
    turquoise: "#40e0d0",
    violet: "#ee82ee",
    wheat: "#f5deb3",
    white: "#ffffff",
    whitesmoke: "#f5f5f5",
    yellow: "#ffff00",
    yellowgreen: "#9acd32",
};
TSColor.__WebColorsHexParsers = [
    null, null, null,
    {
        rx: /^([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])$/,
        short: true
    },
    {
        rx: /^#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])$/,
        short: true
    },
    null,
    {
        rx: /^([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,
        short: false
    },
    {
        rx: /^#([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,
        short: false
    },
    {
        rx: /^([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,
        short: false
    },
    {
        rx: /^#([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/,
        short: false
    }
];
function _isShortRGB(R, G, B) {
    return (R >> 4 & 0x0f) === (R & 0x0f) &&
        (G >> 4 & 0x0f) === (G & 0x0f) &&
        (B >> 4 & 0x0f) === (B & 0x0f);
}
function _colorToShortCSS(R, G, B) {
    return `#${(R & 0x0f).toString(16)}${(G & 0x0f).toString(16)}${(B & 0x0f).toString(16)}`;
}
function _colorToStandardCSS(R, G, B, A) {
    function _toHex(v) {
        const s = v.toString(16);
        return v >= 16 ? s : "0" + s;
    }
    if ((0, commons_1.$ok)(A)) {
        return `#${_toHex(R)}${_toHex(G)}${_toHex(B)}${_toHex(A)}`;
    }
    return `#${_toHex(R)}${_toHex(G)}${_toHex(B)}`;
}
function _component(x) { return isNaN(x) ? 0 : Math.min(1, Math.max(0, x)); }
function _cmykLighter(X) { X = _component(X); return -(X * X) / 3.0 + 5.0 * (X) / 6.0; }
function _cmykDarker(X) { X = _component(X); return 2 * (X * X) / 3.0 + X / 2.0 + 0.25; }
function _rgbLighter(X) { return ((_cmykDarker(X / 255.0) * 255) | 0); }
function _rgbDarker(X) { return ((_cmykLighter(X / 255.0) * 255) | 0); }
//# sourceMappingURL=tscolor.js.map