"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.$arrayBufferFromBuffer = exports.$bufferFromArrayBuffer = exports.$bytesFromAsciiString = exports.TSData = void 0;
const commons_1 = require("./commons");
const fs_1 = require("./fs");
const types_1 = require("./types");
const utils_1 = require("./utils");
;
class TSData {
    constructor(source, opts = {}) {
        this._allocFn = (0, commons_1.$ok)(opts.allocMethod) ? opts.allocMethod : (opts.fillWithZeros ? Buffer.alloc : Buffer.allocUnsafe);
        if (!(0, commons_1.$ok)(source)) {
            source = 0;
        }
        if (source instanceof TSData) {
            this._len = source._len;
            this._buf = this._allocFn(source.capacity);
            if (this._len > 0) {
                source._copyTo(this._buf);
            }
        }
        else if (source instanceof Buffer) {
            // no copy here
            this._len = source.length;
            if (opts.dontCopySourceBuffer) {
                this._buf = source;
            }
            else {
                this._buf = this._allocFn((0, commons_1.$capacityForCount)(this._len));
                source.copy(this._buf);
            }
        }
        else if (source instanceof ArrayBuffer || source instanceof Uint8Array) {
            this._buf = $bufferFromArrayBuffer(source);
            this._len = this._buf.length;
        }
        else if ((0, commons_1.$isunsigned)(source)) {
            const capacity = (0, commons_1.$capacityForCount)(source);
            this._len = 0;
            this._buf = this._allocFn(capacity);
        }
        else {
            throw 'Bad TSData constructor() number parameter';
        }
    }
    static fromFile(src) {
        if ((0, utils_1.$inbrowser)()) {
            throw 'TSData.fromFile(): unavailable static method in browser';
        }
        const b = (0, fs_1.$readBuffer)(src);
        return (0, commons_1.$ok)(b) ? new TSData(b, { dontCopySourceBuffer: true }) : null;
    }
    [Symbol.iterator]() {
        let pos = 0;
        return { next: () => {
                return pos < this._len ? { done: false, value: this._buf[pos++] } : { done: true, value: NaN };
            } };
    }
    clone() { return new TSData(this, { allocMethod: this._allocFn }); }
    appendData(source, start = 0, end) {
        let len = (0, commons_1.$length)(source);
        if (start < 0) {
            start = 0;
        }
        end = (0, commons_1.$ok)(end) && end < len ? end : len;
        if (start < end) {
            this._willGrow(end - start);
            const b = source instanceof TSData ? source._buf : source;
            b.copy(this._buf, this._len, start, end);
            this._len += end - start;
        }
    }
    setData(source, targetStart = 0, sourceStart = 0, sourceEnd) {
        let len = (0, commons_1.$length)(source);
        sourceEnd = (0, commons_1.$ok)(sourceEnd) ? Math.min(sourceEnd, len) : len;
        len = sourceEnd - sourceStart;
        if (len > 0) {
            if (targetStart + len > this._len) {
                this._willGrow(targetStart + len - this._len);
            }
            for (let i = this._len; i < targetStart; i++) {
                this._buf[i] = 0;
            } // fill intermediate part with zeros
            source.copy(this._buf, targetStart, sourceStart, sourceEnd);
            this._len = targetStart + len;
        }
        else {
            this._len = 0;
        }
    }
    appendByte(source) {
        this._willGrow(1);
        this._buf[this._len++] = source;
    }
    appendBytes(source, start = 0, end) {
        let len = (0, commons_1.$count)(source);
        if (start < 0) {
            start = 0;
        }
        end = (0, commons_1.$ok)(end) && end < len ? end : len;
        if (start < end) {
            this._willGrow(end - start);
            for (let i = start; i < end; i++) {
                this._buf[this._len++] = source[i];
            }
        }
    }
    appendAsciiString(source) {
        this.appendBytes($bytesFromAsciiString(source));
    }
    replaceBytes(source, targetStart = 0, sourceStart = 0, sourceEnd) {
        let len = (0, commons_1.$count)(source);
        sourceEnd = (0, commons_1.$ok)(sourceEnd) ? Math.min(sourceEnd, len) : len;
        len = sourceEnd - sourceStart;
        if (len > 0) {
            if (targetStart + len > this._len) {
                this._willGrow(targetStart + len - this._len);
            }
            for (let i = this._len; i < targetStart; i++) {
                this._buf[i] = 0;
            } // fill intermediate part with zeros
            for (let i = 0; i < len; i++) {
                this._buf[targetStart + i] = source[sourceStart + i];
            }
            if (targetStart + len > this._len) {
                this._len = targetStart + len;
            }
        }
    }
    replaceAsciiString(source, targetStart) {
        this.replaceBytes($bytesFromAsciiString(source), targetStart);
    }
    get capacity() { return this._buf.length; }
    get length() { return this._len; }
    set length(n) {
        if (!(0, commons_1.$isunsigned)(n)) {
            throw `TSDate.length = ${n} is not valid.`;
        }
        if (n > this._len) {
            this._willGrow(this._len - n);
            if (this._allocFn !== Buffer.alloc) {
                while (this._len < n) {
                    this._buf[this._len++] = 0;
                }
            }
        }
        else {
            this._len = n;
        }
    }
    get byteLength() { return this._len; }
    get mutableBuffer() { return this._len === this.capacity ? this._buf : this._buf.slice(0, this._len); }
    get buffer() {
        const ret = Buffer.allocUnsafe(this._len);
        this._copyTo(ret);
        return ret;
    }
    get internalStorage() { return [this._buf, this._len]; } // use that to your own risk
    entries() { return this.mutableBuffer.entries(); }
    keys() { return this.mutableBuffer.keys(); }
    values() { return this.mutableBuffer.values(); }
    includes(value, byteOffset = 0) {
        const slen = _searchedLength(value);
        if (slen <= 0 || !(0, commons_1.$isunsigned)(byteOffset) || byteOffset + slen > this._len) {
            return false;
        }
        if (value instanceof TSData) {
            value = value.mutableBuffer;
        }
        return this._buf.includes(value, byteOffset);
    }
    // warning : byteOffset should be >= 0
    indexOf(value, byteOffset = 0) {
        const slen = _searchedLength(value);
        if (slen <= 0 || !(0, commons_1.$isunsigned)(byteOffset) || byteOffset + slen > this._len) {
            return -1;
        }
        if (value instanceof TSData) {
            value = value.mutableBuffer;
        }
        return this._buf.indexOf(value, byteOffset);
    }
    // warning : byteOffset should be >= 0
    lastIndexOf(value, byteOffset = 0) {
        const slen = _searchedLength(value);
        if (slen <= 0 || !(0, commons_1.$isunsigned)(byteOffset) || byteOffset + slen > this._len) {
            return -1;
        }
        if (value instanceof TSData) {
            value = value.mutableBuffer;
        }
        return this._buf.lastIndexOf(value, byteOffset);
    }
    // with slice, you get a new TSData which holds a copy of the sliced data
    slice(begin = 0, end = this._len) {
        if (begin < 0) {
            begin = 0;
        }
        if (end > this._len) {
            end = this._len;
        }
        if (begin >= end) {
            return new TSData(0);
        }
        const finalLen = end - begin;
        const ret = new TSData(finalLen, { allocMethod: this._allocFn });
        this._buf.copy(ret._buf, 0, begin, end);
        ret._len = finalLen;
        return ret;
    }
    copy(targetBuffer, targetStart = 0, sourceStart = 0, sourceEnd = this._len) {
        if (sourceEnd > this._len) {
            sourceEnd = this._len;
        }
        if (sourceStart < sourceEnd) {
            if (targetBuffer instanceof Uint8Array) {
                return this._buf.copy(targetBuffer, targetStart, sourceStart, sourceEnd);
            }
            targetBuffer.replaceBytes(this._buf, targetStart, sourceStart, sourceEnd);
            return sourceEnd - sourceStart;
        }
        return 0;
    }
    writeToFile(path, opts) {
        if ((0, utils_1.$inbrowser)()) {
            throw 'TSData.writeToFile(): unavailable method in browser';
        }
        return (0, fs_1.$writeBuffer)(path, this, opts);
    }
    // second part of the return tupple may contains the path of the precedent version of the data
    fullWriteToFile(path, opts) {
        if ((0, utils_1.$inbrowser)()) {
            throw 'TSData.fullWriteToFile(): unavailable method in browser';
        }
        return (0, fs_1.$fullWriteBuffer)(path, this, opts);
    }
    equals(otherBuffer) { return this.isEqual(otherBuffer); }
    toArrayBuffer() { return $arrayBufferFromBuffer(this._buf, 0, this._len); }
    // ============ TSObject conformance =============== 
    toString(encoding = 'binary', start = 0, end = this._len) {
        start = Math.max(0, start);
        end = Math.min(end, this._len);
        if (start >= end) {
            return '';
        }
        return (0, commons_1.$isfunction)(encoding) ?
            encoding(this._buf, start, end) :
            this._buf.toString(encoding, start, end);
    }
    toJSON() { return this.mutableBuffer.toJSON(); }
    toArray() {
        let ret = [];
        for (let i = 0; i < this._len; i++) {
            ret[i] = this._buf[i];
        }
        return ret;
    }
    compare(other) {
        if (this === other) {
            return types_1.Same;
        }
        if (other instanceof TSData) {
            return Buffer.compare(this.mutableBuffer, other.mutableBuffer);
        }
        if (other instanceof Uint8Array) {
            return Buffer.compare(this.mutableBuffer, other);
        }
        return undefined;
    }
    isEqual(other) {
        if (this === other) {
            return true;
        }
        if (other instanceof TSData) {
            if (other._len !== this._len) {
                return false;
            }
            return Buffer.compare(this.mutableBuffer, other.mutableBuffer) === 0;
        }
        else if (other instanceof Uint8Array || other instanceof Buffer) {
            if (other.length !== this._len) {
                return false;
            }
            return Buffer.compare(this.mutableBuffer, other) === 0;
        }
        return false;
    }
    readBigUInt64BE(offset) { return this._read(offset, 8, Buffer.prototype.readBigUInt64BE); }
    readBigUInt64LE(offset) { return this._read(offset, 8, Buffer.prototype.readBigUInt64LE); }
    readBigInt64BE(offset) { return this._read(offset, 8, Buffer.prototype.readBigInt64BE); }
    readBigInt64LE(offset) { return this._read(offset, 8, Buffer.prototype.readBigInt64LE); }
    readUInt8(offset) { return this._read(offset, 1, Buffer.prototype.readUInt8); }
    readUInt16LE(offset) { return this._read(offset, 2, Buffer.prototype.readUInt16LE); }
    readUInt16BE(offset) { return this._read(offset, 2, Buffer.prototype.readUInt16BE); }
    readUInt32LE(offset) { return this._read(offset, 4, Buffer.prototype.readUInt32LE); }
    readUInt32BE(offset) { return this._read(offset, 4, Buffer.prototype.readUInt32BE); }
    readInt8(offset) { return this._read(offset, 1, Buffer.prototype.readInt8); }
    readInt16LE(offset) { return this._read(offset, 2, Buffer.prototype.readInt16LE); }
    readInt16BE(offset) { return this._read(offset, 2, Buffer.prototype.readInt16BE); }
    readInt32LE(offset) { return this._read(offset, 4, Buffer.prototype.readInt32LE); }
    readInt32BE(offset) { return this._read(offset, 4, Buffer.prototype.readInt32BE); }
    readFloatLE(offset) { return this._read(offset, 4, Buffer.prototype.readFloatLE); }
    readFloatBE(offset) { return this._read(offset, 4, Buffer.prototype.readFloatBE); }
    readDoubleLE(offset) { return this._read(offset, 8, Buffer.prototype.readDoubleLE); }
    readDoubleBE(offset) { return this._read(offset, 8, Buffer.prototype.readDoubleBE); }
    writeBigInt64BE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeBigInt64BE); }
    writeBigInt64LE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeBigInt64LE); }
    writeBigUInt64BE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeBigUInt64BE); }
    writeBigUInt64LE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeBigUInt64LE); }
    writeUInt8(value, offset) { this._write(value, offset, 1, Buffer.prototype.writeUInt8); }
    writeUInt16LE(value, offset) { this._write(value, offset, 2, Buffer.prototype.writeUInt16LE); }
    writeUInt16BE(value, offset) { this._write(value, offset, 2, Buffer.prototype.writeUInt16BE); }
    writeUInt32LE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeUInt32LE); }
    writeUInt32BE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeUInt32BE); }
    writeInt8(value, offset) { this._write(value, offset, 1, Buffer.prototype.writeInt8); }
    writeInt16LE(value, offset) { this._write(value, offset, 2, Buffer.prototype.writeInt16LE); }
    writeInt16BE(value, offset) { this._write(value, offset, 2, Buffer.prototype.writeInt16BE); }
    writeInt32LE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeInt32LE); }
    writeInt32BE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeInt32BE); }
    writeFloatLE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeFloatLE); }
    writeFloatBE(value, offset) { this._write(value, offset, 4, Buffer.prototype.writeFloatBE); }
    writeDoubleLE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeDoubleLE); }
    writeDoubleBE(value, offset) { this._write(value, offset, 8, Buffer.prototype.writeDoubleBE); }
    // ============ private methods =============== 
    _willGrow(n) {
        if (n && this._len + n > this.capacity) {
            const newCapacity = (0, commons_1.$capacityForCount)((this._len + n));
            let newBuffer = this._allocFn(newCapacity);
            this._copyTo(newBuffer);
            this._buf = newBuffer;
        }
    }
    _copyTo(target) {
        if (this._len > 0) {
            this._buf.copy(target, 0, 0, this._len);
        }
    }
    _read(offset = 0, size, bufferReadFn) {
        if (!(0, commons_1.$isunsigned)(offset) || offset + size > this._len) {
            throw `TSData._read(${offset}) out of bound [0,${this._len}]`;
        }
        return bufferReadFn.call(this._buf, offset);
    }
    _write(value, offset = 0, size, bufferWriteFn) {
        if (!(0, commons_1.$isunsigned)(offset)) {
            throw `TSData._write(${value}, ${offset}) out of bound [0,${this._len}]`;
        }
        if (offset + size > this._len) {
            this._willGrow(offset + size - this._len);
        }
        for (let i = this._len; i < offset; i++) {
            this._buf[i] = 0;
        } // fill intermediate part with zeros
        bufferWriteFn.call(this._buf, value, offset);
    }
}
exports.TSData = TSData;
function _searchedLength(value) {
    if ((0, commons_1.$isnumber)(value)) {
        return (0, commons_1.$isunsigned)(value, types_1.UINT8_MAX) ? 1 : -1;
    }
    return (0, commons_1.$ok)(value) ? value.length : -1;
}
function $bytesFromAsciiString(source, start = 0, end = (0, commons_1.$length)(source)) {
    let bytes = [];
    if (!(0, commons_1.$isunsigned)(start) || !(0, commons_1.$isunsigned)(end)) {
        throw '$bytesFromAsciiString(): start and end parameters must be true unsigned values';
    }
    end = Math.min((0, commons_1.$length)(source), (0, commons_1.$tounsigned)(end));
    start = Math.min(end, (0, commons_1.$tounsigned)(start));
    for (let i = start, j = 0; i < end; i++, j++) {
        const c = source.charCodeAt(i);
        if (c < 128) {
            bytes[j] = c;
        }
    }
    return bytes;
}
exports.$bytesFromAsciiString = $bytesFromAsciiString;
function $bufferFromArrayBuffer(a) {
    return ArrayBuffer.isView(a) ? Buffer.from(a.buffer, a.byteOffset, a.byteLength) : Buffer.from(a);
}
exports.$bufferFromArrayBuffer = $bufferFromArrayBuffer;
function $arrayBufferFromBuffer(source, start = 0, end = source.length) {
    if (!(0, commons_1.$isunsigned)(start) || !(0, commons_1.$isunsigned)(end)) {
        throw '$arrayBufferFromBuffer(): start and end parameters must be true unsigned values';
    }
    end = Math.min(source.length, (0, commons_1.$tounsigned)(end));
    start = Math.min(end, (0, commons_1.$tounsigned)(start));
    const ret = new ArrayBuffer(end - start);
    const view = new Uint8Array(ret);
    for (let i = start, j = 0; i < end; i++, j++) {
        view[j] = source[i];
    }
    return ret;
}
exports.$arrayBufferFromBuffer = $arrayBufferFromBuffer;
//# sourceMappingURL=tsdata.js.map