"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSInterval = void 0;
const commons_1 = require("./commons");
const tsrange_1 = require("./tsrange");
const tsdate_1 = require("./tsdate");
const types_1 = require("./types");
const tscouple_1 = require("./tscouple");
/************************************************
 *  WARNING: range conformance methods and containing
 *  and intersectionning methods do round dates to minutes
 ************************************************/
class TSInterval extends tscouple_1.TSCouple {
    static make(a, b) { return new TSInterval(a, b); }
    get start() { return (0, commons_1.$ok)(this.first) ? this.first : new tsdate_1.TSDate(tsdate_1.TSDate.PAST); }
    get end() { return (0, commons_1.$ok)(this.second) ? this.second : new tsdate_1.TSDate(tsdate_1.TSDate.FUTURE); }
    clone() { return new TSInterval(this.first, this.second); }
    get hasSignificantRange() {
        if ((0, commons_1.$ok)(this.first) && (0, commons_1.$ok)(this.second)) {
            const s = this.start.timestamp;
            const e = this.end.timestamp;
            return (0, commons_1.$isint)(s) && (0, commons_1.$isint)(e) && s < e;
        }
        return false;
    }
    get isValid() { return (0, commons_1.$ok)(this.first); }
    get isEmpty() { return (0, commons_1.$ok)(this.first) && (0, commons_1.$ok)(this.second) && this.start.timestamp === this.end.timestamp; }
    get range() { return this.hasSignificantRange ? new tsrange_1.TSRange(this.start, this.end) : (0, tsrange_1.TSBadRange)(); }
    daysInterval() {
        if ((0, commons_1.$ok)(this.first)) {
            if ((0, commons_1.$ok)(this.second)) {
                let s = this.start.dateWithoutTime();
                if (this.isEmpty) {
                    return new TSInterval(s, s);
                }
                let e = this.end.dateWithoutTime();
                if (s.isEqual(e)) {
                    e = e.dateByAdding(0, 0, 1);
                }
                return new TSInterval(s, e);
            }
            return new TSInterval(this.start.dateWithoutTime(), null);
        }
        return (0, commons_1.$ok)(this.second) ? new TSInterval(null, this.end.dateWithoutTime()) : new TSInterval(null, null);
    }
    hasSameRange(other) {
        return this.hasSignificantRange && other.hasSignificantRange && this.range.isEqual(other.range);
        // infinite intervals never have the same range
    }
    containsDate(aDate) {
        if (this.isEmpty) {
            return false;
        } // this is not an interval
        let loc = aDate.timestamp;
        if (this.hasSignificantRange) {
            return this.range.containsLocation(loc);
        }
        else if ((0, commons_1.$ok)(this.first)) {
            return this.start.timestamp >= loc;
        }
        else if ((0, commons_1.$ok)(this.second)) {
            return loc < this.end.timestamp;
        }
        return true; // a both way infinite interval always contains any date
    }
    intersects(other) {
        if (this.isEmpty || other.isEmpty) {
            return false;
        } // none of these objects is an interval
        else if (this.hasSignificantRange) {
            if (other.hasSignificantRange) {
                return this.range.intersects(other.range);
            }
            else if ((0, commons_1.$ok)(other.first)) {
                return other.start.timestamp < this.end.timestamp;
            }
            else if ((0, commons_1.$ok)(other.second)) {
                return other.end.timestamp > this.start.timestamp;
            }
            return true; // intersection with a both way infinite interval is always true
        }
        else if (other.hasSignificantRange) {
            return other.intersects(this);
        }
        else if ((0, commons_1.$ok)(this.first)) {
            return (0, commons_1.$ok)(other.second) ? other.end.timestamp > this.start.timestamp : true;
        }
        else if ((0, commons_1.$ok)(this.second)) {
            return (0, commons_1.$ok)(other.first) ? this.end.timestamp > other.start.timestamp : true;
        }
        return true; // intersection with a both way infinite interval is always true
    }
    contains(other) {
        if (this.isEmpty || other.isEmpty) {
            return false;
        } // none of these objects is an interval
        else if (this.hasSignificantRange) {
            return other.hasSignificantRange && this.range.contains(other.range);
        }
        else if ((0, commons_1.$ok)(this.first)) {
            return (0, commons_1.$ok)(other.first) && this.start.timestamp <= other.start.timestamp;
        }
        else if ((0, commons_1.$ok)(this.second)) {
            return (0, commons_1.$ok)(other.second) && other.end.timestamp <= this.end.timestamp;
        }
        return true; // a both ways infinite interval always contains any intervals
    }
    containedIn(other) { return other.contains(this); }
    continuousWith(other) {
        return (this.intersects(other) ||
            ((0, commons_1.$ok)(this.end) && (0, commons_1.$ok)(other.start) && this.end.isEqual(other.start)) ||
            ((0, commons_1.$ok)(this.start) && (0, commons_1.$ok)(other.end) && this.start.isEqual(other.end)));
    }
    compare(other) {
        if (this.isEqual(other)) {
            return types_1.Same;
        }
        if (other instanceof TSInterval) {
            if (this.hasSignificantRange) {
                if ((0, commons_1.$ok)(other.first) && this.end.timestamp <= other.start.timestamp) {
                    return types_1.Ascending;
                }
                if ((0, commons_1.$ok)(other.second) && other.end.timestamp <= this.start.timestamp) {
                    return types_1.Descending;
                }
            }
            else if (other.hasSignificantRange) {
                if ((0, commons_1.$ok)(this.first) && other.end.timestamp <= this.start.timestamp) {
                    return types_1.Descending;
                }
                if ((0, commons_1.$ok)(this.second) && this.end.timestamp <= other.start.timestamp) {
                    return types_1.Ascending;
                }
            }
            else if ((0, commons_1.$ok)(this.first) && (0, commons_1.$ok)(other.second) &&
                other.end.timestamp <= this.start.timestamp) {
                return types_1.Descending;
            }
            else if ((0, commons_1.$ok)(this.second) && (0, commons_1.$ok)(other.first) &&
                this.end.timestamp <= other.start.timestamp) {
                return types_1.Ascending;
            }
        }
        return undefined;
    }
    // ============ TSObject conformance =============== 
    isEqual(other) {
        if (this === other) {
            return true;
        }
        return (other instanceof TSInterval) &&
            (((0, commons_1.$ok)(this.first) && this.first.isEqual(other.first)) || (!(0, commons_1.$ok)(this.first) && !(0, commons_1.$ok)(other.first))) &&
            (((0, commons_1.$ok)(this.second) && this.second.isEqual(other.second)) || (!(0, commons_1.$ok)(this.second) && !(0, commons_1.$ok)(other.second)));
    }
    toJSON() { return { start: (0, commons_1.$jsonobj)(this.first), end: (0, commons_1.$jsonobj)(this.second) }; }
    toArray() { return [this]; } // should we not return one or two dates here ?
}
exports.TSInterval = TSInterval;
//# sourceMappingURL=tsinterval.js.map