"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSList = exports.TSListNode = void 0;
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const types_1 = require("./types");
class TSListNode {
    constructor(data) {
        this.data = data;
        this.next = null;
        this.prev = null;
    }
    // ============ TSObject conformance =============== 
    isEqual(other) {
        return this === other || (other instanceof TSListNode && (0, compare_1.$equal)(other.data, this.data));
    }
    compare(other) {
        if (this === other) {
            return types_1.Same;
        }
        return (other instanceof TSListNode) ? (0, compare_1.$compare)(this.data, other.data) : undefined;
    }
    toJSON() { return (0, commons_1.$jsonobj)(this.data); }
    toString() { return (0, commons_1.$string)(this.data); }
    toArray() { return [this]; }
}
exports.TSListNode = TSListNode;
class TSList {
    constructor() {
        this._f = null;
        this._l = null;
        this._n = 0;
    }
    get length() { return this._n; }
    get count() { return this._n; }
    get first() { return this._f; }
    get last() { return this._l; }
    insert(data, before) {
        const node = new TSListNode(data);
        if (!this._f) {
            if ((0, commons_1.$ok)(before))
                throw 'Try to insert before a non existing node';
            this._f = this._l = node;
            this._n = 1;
        }
        else if (!(0, commons_1.$ok)(before) || before === this._f) {
            this._f.prev = node;
            node.next = this._f;
            this._f = node;
            node.prev = null;
            this._n++;
        }
        /* warning, we dont verify the node appartenance here */
        else {
            const prev = before.prev;
            prev.next = node;
            node.prev = prev;
            node.next = before;
            before.prev = node;
            this._n++;
        }
        return node;
    }
    add(data) {
        if (!this._f)
            return this.insert(data);
        const node = new TSListNode(data);
        this._l.next = node;
        node.prev = this._l;
        this._l = node;
        this._n++;
        return node;
    }
    /* warning, we dont verify the node appartenance here */
    removeNode(node) {
        if (node === this._f) {
            this._f = node.next;
            if (this._f)
                this._f.prev = null;
            else
                this._l = null;
        }
        else if (node === this._l) {
            this._l = node.prev;
            if (this._l)
                this._l.next = null;
            else
                this._f = null; // should never occur
        }
        else {
            const previous = node.prev;
            const succ = node.next;
            previous.next = succ;
            succ.prev = previous;
        }
        this._n--;
        node.next = node.prev = null; // remove useless pointers
    }
    clear() {
        let l = this._f;
        if (l != null) {
            while (l.next !== null) {
                l.prev = null;
                l = l.next;
                l.prev.next = null;
            }
            this._f = this._l = null;
            this._n = 0;
        }
    }
    forEach(callback) {
        if (this._f) {
            const traverseMe = (node) => {
                callback(node.data);
                if (node.next)
                    traverseMe(node.next);
            };
            traverseMe(this._f);
        }
    }
    [Symbol.iterator]() {
        let currentNode = this._f;
        return { next: () => {
                if (currentNode) {
                    const ret = { done: false, value: currentNode.data };
                    currentNode = currentNode.next;
                    return ret;
                }
                return { done: true, value: undefined };
            } };
    }
    search(callback) {
        if (this._f) {
            const findMe = (node) => {
                if (callback(node.data)) {
                    return node;
                }
                return node.next ? findMe(node.next) : null;
            };
            return findMe(this._f);
        }
        return null;
    }
    // ============ TSClone conformance =============== 
    clone() {
        let copy = new TSList();
        let node = this._f;
        while (node) {
            copy.add(node.data);
            node = node.next;
        }
        return copy;
    }
    // ============ TSObject conformance =============== 
    isEqual(other) {
        if (this === other) {
            return true;
        }
        if (!(other instanceof TSList) || this._n !== other.count) {
            return false;
        }
        let a = this._f;
        let b = other.first;
        while (a !== null && b !== null) {
            if (!a.isEqual(b)) {
                return false;
            }
            a = a.next;
            b = b.next;
        }
        return a === null && b === null;
    }
    compare(other) {
        if (this === other) {
            return types_1.Same;
        }
        if (!(other instanceof TSList)) {
            return undefined;
        }
        let a = this._f;
        let b = other.first;
        while (a !== null && b !== null) {
            const c = a.compare(b);
            if (c !== types_1.Same) {
                return c;
            }
            a = a.next;
            b = b.next;
        }
        return a === null ? (b === null ? types_1.Same : types_1.Ascending) : types_1.Descending;
    }
    toString(opts = { prefix: '(', separator: ',', suffix: ')' }) {
        let s = (0, commons_1.$ok)(opts.prefix) ? opts.prefix : '';
        if (!(0, commons_1.$ok)(opts.printer))
            opts.printer = d => d.toString();
        if (this._f) {
            const hasSepa = (0, commons_1.$length)(opts.separator) > 0;
            const printMe = (node) => {
                const element = opts.printer(node.data);
                if ((0, commons_1.$ok)(element)) {
                    s = s + element;
                }
                if (node.next !== null) {
                    if (hasSepa) {
                        s = s + opts.suffix;
                    }
                    printMe(node.next);
                }
            };
            printMe(this._f);
        }
        if ((0, commons_1.$length)(opts.suffix))
            s = s + opts.suffix;
        return s;
    }
    toJSON() { return this.toArray(e => (0, commons_1.$jsonobj)(e)); }
    toArray(map) {
        let array = [];
        if (this._f) {
            const addMe = (node) => {
                if (map) {
                    const v = map(node.data);
                    if ((0, commons_1.$ok)(v))
                        array.push(v);
                }
                else
                    array.push(node.data);
                if (node.next)
                    addMe(node.next);
            };
            addMe(this._f);
        }
        return array;
    }
    // ============ TSCollection conformance =============== 
    getItems() { return this.toArray(); }
}
exports.TSList = TSList;
//# sourceMappingURL=tslist.js.map