"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSServerLogType = exports.TSServer = exports.TSParametricTokenType = void 0;
const http_1 = require("http");
const commons_1 = require("./commons");
const tserrors_1 = require("./tserrors");
const tsrequest_1 = require("./tsrequest");
const types_1 = require("./types");
const utils_1 = require("./utils");
const tsservercomp_1 = require("./tsservercomp");
exports.TSParametricTokenType = {
    string: 'string',
    number: 'number',
    int: 'int',
    unsigned: 'unsigned',
    boolean: 'boolean',
    date: 'date',
    color: 'color',
    uuid: 'uuid',
    email: 'email'
};
class TSServer {
    // =================== CONSTRUCTOR =======================
    constructor(endPoints, opts = {}) {
        this._logger = (0, commons_1.$ok)(opts.logger) ? opts.logger : _internalLogger;
        this._logInfos = !!opts.logInfos;
        // ========= first construct the static websites architecture (only if we're not inside a browser )==========
        this._sites = [];
        if ((0, commons_1.$ok)(opts.webSites)) {
            const keys = (0, commons_1.$keys)(opts.webSites);
            if ((0, utils_1.$inbrowser)() && keys.length) {
                throw 'TSServer cannot handle static websides inside a browser';
            }
            keys.forEach(u => {
                this._sites.push(new tsservercomp_1.TSStaticWebsite(u, opts.webSites[u], {
                    logger: this._logger,
                    managedTypes: opts.managedTypes,
                    maxCacheSize: opts.maxCacheSize,
                    maxCachedFileSize: opts.maxCachedFileSize,
                    maxCachedFiles: opts.maxCachedFiles,
                    maxBlacklistedFiles: opts.maxBlacklistedFiles
                }));
            });
        }
        // ========= second construct the dynamic endpoints ==========
        this._endPoints = [];
        (0, commons_1.$keys)(endPoints).forEach(path => {
            this._endPoints.push(new tsservercomp_1.TSParametricEndPoints(path, endPoints[path], true));
        });
        if ((0, commons_1.$ok)(opts.port)) {
            if (!(0, commons_1.$isunsigned)(opts.port, types_1.UINT16_MAX)) {
                throw `Bad HTTP server port ${opts.port}.`;
            }
            this.port = opts.port;
        }
        else {
            this.port = 3000;
        }
        this.host = (0, commons_1.$ftrim)(opts.host);
        if (!this.host.length) {
            this.host = 'http://localhost/';
        }
    }
    // =================== static methods =======================
    static start(endPoints, opts) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!TSServer.__server) {
                TSServer.__server = new TSServer(endPoints, opts);
                yield TSServer.__server._start();
            }
        });
    }
    static clearCaches() {
        return __awaiter(this, void 0, void 0, function* () { if (TSServer.__server) {
            yield TSServer.__server._clearCaches();
        } });
    }
    // returns undefined if the server is stoped.
    // after that you can do a new server start
    static stop() {
        return __awaiter(this, void 0, void 0, function* () {
            if (TSServer.__server) {
                const e = yield TSServer.__server._stop();
                if (!(0, commons_1.$ok)(e)) {
                    delete TSServer.__server;
                }
                return e;
            }
            return undefined;
        });
    }
    // =================== instance methods =======================
    _start() {
        return __awaiter(this, void 0, void 0, function* () {
            this._httpServer = (0, http_1.createServer)((req, res) => __awaiter(this, void 0, void 0, function* () {
                try {
                    // validating method
                    const method = tsservercomp_1.TSParametricEndPoints.validRequestMethod(req.method);
                    const url = new URL((0, commons_1.$string)(req.url), this.host);
                    if (!(0, commons_1.$ok)(method)) {
                        throw new tserrors_1.TSHttpError(`Request method '${req.method}' is not allowed.`, tsrequest_1.Resp.NotAllowed, {
                            method: req.method,
                            path: (0, commons_1.$length)(url.pathname) ? url.pathname : '/'
                        });
                    }
                    // validating url
                    if (!(0, commons_1.$length)(url.pathname) || url.pathname === '/') {
                        throw new tserrors_1.TSHttpError('Root path is not Accessible', tsrequest_1.Resp.Forbidden, {
                            method: req.method,
                            path: '/'
                        });
                    }
                    let pep = undefined;
                    let parameters = {};
                    this._endPoints.forEach(ep => {
                        const params = ep.parametersFromPath(url.pathname);
                        if ((0, commons_1.$ok)(params) && (!(0, commons_1.$defined)(pep) || pep.depth < ep.depth)) {
                            pep = ep;
                            parameters = params;
                        }
                    });
                    if ((0, commons_1.$ok)(pep)) {
                        // we have a potential dynamic resource ;
                        yield pep.execute({
                            url: url,
                            method: method,
                            parameters: parameters,
                            message: req,
                            query: {} // the final query will be calculated after
                        }, res);
                        if (this._logInfos) {
                            yield this._logger(this, req, TSServerLogType.Log, `did handle resource '${url.pathname}'.`);
                        }
                        return;
                    }
                    else if (method === tsrequest_1.Verb.Get) {
                        for (let s of this._sites) {
                            const [b, type] = s.getStaticResource(url.pathname);
                            if ((0, commons_1.$ok)(b)) {
                                if (this._logInfos) {
                                    yield this._logger(this, req, TSServerLogType.Log, `did handle static resource '${url.pathname}'.`);
                                }
                                res.setHeader('Content-Type', type);
                                res.writeHead(tsrequest_1.Resp.OK);
                                res.end(b);
                                return;
                            }
                        }
                    }
                    // here the resource is not found
                    throw new tserrors_1.TSHttpError(`endpoint ${method} '${url.pathname}' was not found.`, tsrequest_1.Resp.NotFound, {
                        method: method,
                        path: (0, commons_1.$length)(url.pathname) ? url.pathname : '/'
                    });
                }
                catch (e) {
                    let ret = {};
                    ret.status = (0, commons_1.$isunsigned)(e === null || e === void 0 ? void 0 : e.status) && Object.values(tsrequest_1.Resp).includes(e.status) ? e.status : tsrequest_1.Resp.InternalError;
                    ret.error = e.message;
                    if (!(0, commons_1.$length)(ret.error)) {
                        ret.error = 'Unknown internal Error';
                    }
                    ;
                    if ((0, commons_1.$ok)(e.infos)) {
                        ret.infos = e.infos;
                    }
                    yield this._logger(this, req, TSServerLogType.Warning, `${ret.status} - ${ret.error}`);
                    res.writeHead(ret.status, { 'Content-Type': 'application/json' });
                    res.end(JSON.stringify(ret));
                }
            }));
            this._httpServer.listen(this.port, () => __awaiter(this, void 0, void 0, function* () {
                yield this._logger(this, undefined, TSServerLogType.Log, `running on port ${this.port} '${this.host}' ...`);
            }));
        });
    }
    _clearCaches() {
        return __awaiter(this, void 0, void 0, function* () { for (let s of this._sites) {
            yield s.clearCaches();
        } });
    }
    _stop() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._logger(this, undefined, TSServerLogType.Log, `server is exiting...`);
            if ((0, commons_1.$ok)(this._httpServer)) {
                const ret = yield _internalStopServer(this._httpServer);
                if ((0, commons_1.$ok)(ret)) {
                    yield this._logger(this, undefined, TSServerLogType.Error, `cannot stop for reason ${ret.name}:\n${ret.message}`);
                }
                return ret;
            }
            ;
            return undefined;
        });
    }
}
exports.TSServer = TSServer;
TSServer.__server = undefined;
var TSServerLogType;
(function (TSServerLogType) {
    TSServerLogType["Log"] = "Info";
    TSServerLogType["Warning"] = "Warning";
    TSServerLogType["Error"] = "Error";
})(TSServerLogType = exports.TSServerLogType || (exports.TSServerLogType = {}));
;
// ================ private functions =======================
const _internalStopServer = (server) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((resolve, reject) => {
        server.close((error) => {
            if ((0, commons_1.$ok)(error)) {
                return reject(error);
            }
            resolve(undefined);
        });
    });
});
const _internalLogger = (server, _, type, message) => __awaiter(void 0, void 0, void 0, function* () {
    (0, utils_1.$logterm)(`&0&xfoundation-ts[&wminimal server&x:${server.port}]-` + __TSServerlogHeaders[type] + "&0 &w" + message + "&0");
});
const __TSServerlogHeaders = {
    'Info': "&C&w LOGGING ",
    'Warning': "&O&w WARNING ",
    'Error': "&R&w  ERROR  "
};
//# sourceMappingURL=tsserver.js.map