"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSParametricEndPoints = exports.TSStaticWebsite = void 0;
const fs_1 = require("./fs");
const commons_1 = require("./commons");
const tsrequest_1 = require("./tsrequest");
const tserrors_1 = require("./tserrors");
const tsserver_1 = require("./tsserver");
const tsdate_1 = require("./tsdate");
const tscolor_1 = require("./tscolor");
class TSStaticWebsite {
    constructor(uri, folder, opts) {
        if (!(0, commons_1.$length)(uri)) {
            throw `Static website url not defined for folder '${folder}'`;
        }
        if (!(0, fs_1.$isdirectory)(folder)) {
            throw `website ${uri} root folder ${folder} was not found.`;
        }
        this.uri = uri.toLowerCase();
        this.folder = folder;
        this._opts = Object.assign({}, opts);
        this._cache = new Map();
        this._cacheMemory = 0;
        this._blacklisted = new Set();
        if (!(0, commons_1.$ok)(this._opts.logger)) {
            throw 'You must specify a logger to a new TSStaticWebsite';
        }
        if (!(0, commons_1.$objectcount)(this._opts.managedTypes)) {
            this._opts.managedTypes = Object.assign({}, TSStaticWebsite.__TSServerStandardsTypes);
        }
        if (!(0, commons_1.$ok)(this._opts.maxCacheSize)) {
            this._opts.maxCacheSize = 128 * 1024 * 1024;
        }
        if (!(0, commons_1.$ok)(this._opts.maxCachedFiles)) {
            this._opts.maxCachedFiles = 10000;
        }
        if (!(0, commons_1.$ok)(this._opts.maxCachedFileSize)) {
            this._opts.maxCachedFileSize = 256 * 1024;
        }
        if (!(0, commons_1.$ok)(this._opts.maxBlacklistedFiles)) {
            this._opts.maxBlacklistedFiles = 50000;
        }
    }
    getStaticResource(uri) {
        let ret = undefined;
        let type = '';
        const lcUri = uri.toLowerCase();
        if (lcUri.startsWith(this.uri)) {
            const ext = (0, fs_1.$ext)(lcUri);
            type = (0, commons_1.$length)(ext) ? this._opts.managedTypes[ext] : '';
            if ((0, commons_1.$length)(type)) {
                const path = uri.slice(this.uri.length);
                if (!this._blacklisted.has(path)) {
                    ret = this._cache.get(path);
                    if (!(0, commons_1.$ok)(ret)) {
                        const file = (0, fs_1.$path)(this.folder, path);
                        if ((0, fs_1.$isfile)(file)) {
                            ret = (0, fs_1.$readBuffer)(file);
                            if ((0, commons_1.$ok)(ret)) {
                                this._cacheContent(path, ret);
                            }
                            else {
                                this._blacklist(path);
                            }
                        }
                        else {
                            this._blacklist(path);
                        }
                    }
                }
            }
        }
        return [ret, type];
    }
    clearCaches() {
        return __awaiter(this, void 0, void 0, function* () {
            this._cacheMemory = 0;
            this._cache.clear();
            this._blacklisted.clear();
        });
    }
    _cacheContent(path, b) {
        if (this._cache.size < this._opts.maxCachedFiles && b.length <= this._opts.maxCachedFileSize && this._cacheMemory + b.length <= this._opts.maxCacheSize) {
            this._cache.set(path, b);
            this._cacheMemory += b.length;
        }
    }
    _blacklist(path) {
        if (this._blacklisted.size < this._opts.maxBlacklistedFiles) {
            this._blacklisted.add(path);
        }
    }
}
exports.TSStaticWebsite = TSStaticWebsite;
TSStaticWebsite.__TSServerStandardsTypes = {
    'html': 'text/html',
    'htm': 'text/html',
    'css': 'text/css',
    'js': 'text/javascript',
    'svg': 'image/svg+xml',
    'svgz': 'image/svg+xml',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpg',
    'png': 'image/png',
    'gif': 'image/gif',
    'ico': 'image/x-icon',
    'pdf': 'application/pdf'
};
const TSParametersConversions = {
    string: (s) => s,
    number: (s) => Number(s),
    int: (s) => (0, commons_1.$intornull)(s),
    unsigned: (s) => (0, commons_1.$unsignedornull)(s),
    boolean: (s) => {
        s = (0, commons_1.$ftrim)(s).toLowerCase();
        if (s === '1' || s === 'true' || s === 'y' || s === 'yes') {
            return true;
        }
        if (s === '0' || s === 'false' || s === 'n' || s === 'no') {
            return false;
        }
        return undefined;
    },
    date: (s) => tsdate_1.TSDate.fromIsoString(s),
    color: (s) => {
        let c;
        try {
            c = tscolor_1.TSColor.rgb(s);
        }
        catch (_a) {
            c = undefined;
        }
        return c;
    },
    uuid: (s) => (0, commons_1.$UUID)(s),
    email: (s) => (0, commons_1.$email)(s)
};
class TSParametricEndPoints {
    constructor(path, ep, caseInsensitive = false) {
        path = (0, commons_1.$ftrim)(path);
        const len = path.length;
        if ((0, commons_1.$isfunction)(ep)) {
            ep = { GET: { manager: ep } };
        }
        else if ('manager' in ep) {
            ep = { GET: ep };
        }
        if (len < 2) {
            throw `End points path '${path}' is too short`;
        }
        const methods = (0, commons_1.$keys)(ep);
        if (methods.length === 0) {
            throw `End points path '${path}' has no method defined.`;
        }
        this.endPoints = {};
        methods.forEach(m => {
            if (!(0, commons_1.$ok)(TSParametricEndPoints.validRequestMethod(m))) {
                throw `End points path '${path}' did define invalid '${m}' request method.`;
            }
            const v = ep[m];
            let def = (0, commons_1.$isfunction)(v) ? { manager: v } : v;
            let newQuery = {};
            (0, commons_1.$keys)(def.query).forEach(name => {
                const n = (0, commons_1.$ftrim)((0, commons_1.$string)(name));
                if (!n.length) {
                    throw `End points ${m} '${path}' did define an unamed query variable.`;
                }
                else if (n !== n.ascii()) {
                    throw `End points ${m} '${path}' did define an invalid '${name}' query variable.`;
                }
                const q0 = def.query[name];
                const q = (0, commons_1.$isstring)(q0) ? { type: q0 } : q0;
                newQuery[n.toLowerCase()] = q;
            });
            def.query = newQuery; // can be an empty one
            this.endPoints[m] = def;
        });
        this.caseInsensitive = caseInsensitive;
        // ================= automat initialization ===============
        this.uri = '/';
        this.tokens = [];
        this.depth = 1;
        let State;
        (function (State) {
            State[State["Start"] = 0] = "Start";
            State[State["Standard"] = 1] = "Standard";
            State[State["Bracket"] = 2] = "Bracket";
            State[State["Token"] = 3] = "Token";
            State[State["TokenType"] = 4] = "TokenType";
        })(State || (State = {}));
        ;
        let state = State.Start;
        let regString = '';
        let currentToken = '';
        let currentType = '';
        let constructUri = true;
        // ================= automated path analysis ===============
        for (let i = 0; i < len; i++) {
            const c = path.charAt(i);
            //console.log(`state = ${state}, char = "${c}"`) ;
            switch (state) {
                case State.Start:
                    if (c !== '/') {
                        throw `End points path '${path}' is not absolute.`;
                    }
                    state = State.Standard;
                    break;
                case State.Standard:
                    switch (c) {
                        case '{':
                            state = State.Bracket;
                            constructUri = false;
                            break;
                        case '}': throw `Misplaced '}' character in path '${path}'.`;
                        case '/':
                            if (constructUri) {
                                this.uri += c;
                            }
                            else {
                                regString += '\\/';
                            }
                            this.depth++;
                            break;
                        case '-':
                        case '$':
                        case '+':
                        case '*':
                        case '?':
                        case '.':
                        case '(':
                        case ')':
                            if (constructUri) {
                                this.uri += c;
                            }
                            else {
                                regString += '\\' + c;
                            }
                            break;
                        case '^':
                        case '\\':
                        case '[':
                        case ']':
                        case '`':
                        case '|':
                            throw `Found forbidden character '${c}' in path '${path}'.`;
                        default:
                            if (constructUri) {
                                this.uri += caseInsensitive ? c.toLowerCase() : c;
                            }
                            else {
                                regString += caseInsensitive ? c.toLowerCase() : c;
                            }
                            break;
                    }
                    break;
                case State.Bracket:
                    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')) {
                        currentToken = c;
                        state = State.Token;
                        break;
                    }
                    throw `Found forbidden first character '${c}' in parametric token in path '${path}'.`;
                case State.Token:
                    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c === '-' || c === '_' || c === '.') {
                        currentToken += c;
                        break;
                    }
                    else if (c === ':') {
                        // we have a token type here.
                        state = State.TokenType;
                        currentType = '';
                        break;
                    }
                    else if (c === '}') {
                        // default token type is string
                        this.tokens.push({ name: currentToken, type: tsserver_1.TSParametricTokenType.string });
                        regString += '([a-zA-Z][.a-zA-Z0-9_\-]*)';
                        currentToken = '';
                        currentType = '';
                        state = State.Standard;
                        break;
                    }
                    throw `Found forbidden character '${c}' in parametric token in path '${path}'.`;
                case State.TokenType:
                    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')) {
                        currentType += c.toLowerCase();
                        break;
                    }
                    else if (c === '}') {
                        if (Object.values(tsserver_1.TSParametricTokenType).includes(currentType)) {
                            this.tokens.push({ name: currentToken, type: currentType });
                            regString += '([a-zA-Z][.a-zA-Z0-9_\-]*)';
                            currentToken = '';
                            currentType = '';
                            state = State.Standard;
                            break;
                        }
                        throw `Found unknown type '${currentType}' of token '${currentToken}' in path '${path}'.`;
                    }
                    throw `Found forbidden first character '${c}' in parametric type of token '${currentToken}' in path '${path}'.`;
            }
        }
        if (state !== State.Standard) {
            throw `Malformed parametric path '${path}'.`;
        }
        this._regex = regString.length === 0 ? null : (caseInsensitive ? new RegExp('^' + regString + '$', 'i') : new RegExp('^' + regString + '$'));
        this._tlen = this.tokens.length;
    }
    static validRequestMethod(method) {
        if ((0, commons_1.$ok)(method)) {
            method = method.toUpperCase();
            if (TSParametricEndPoints.__verbs.includes(method)) {
                return method;
            }
        }
        return null;
    }
    get static() { return !(0, commons_1.$ok)(this._regex); }
    execute(req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            const ep = this.endPoints[req.method];
            if (!(0, commons_1.$ok)(ep)) {
                throw new tserrors_1.TSHttpError(`Method '${req.method}' not implemented on url '${req.url.pathname}'`, tsrequest_1.Resp.NotImplemented, {
                    method: req.method,
                    path: (0, commons_1.$length)(req.url.pathname) ? req.url.pathname : '/'
                });
            }
            if ((0, commons_1.$ok)(ep.query)) {
                _calculateQuery(req, ep.query);
            }
            yield ep.manager(req, res); // QUESTION: this method may also throw, so get the stack in the infos ?
        });
    }
    // can return an empty string dictionary which means that our path is not parametric
    parametersFromPath(path) {
        if (this.caseInsensitive) {
            path = path.toLowerCase();
        }
        if (path.startsWith(this.uri)) {
            const rp = path.slice(this.uri.length);
            const ret = {};
            if (!rp.length && !(0, commons_1.$ok)(this._regex)) {
                return ret;
            }
            else if (rp.length && (0, commons_1.$ok)(this._regex)) {
                const m = rp.match(this._regex);
                if ((0, commons_1.$ok)(m)) {
                    if (m.length === this._tlen + 1) {
                        for (let i = 0; i < this._tlen; i++) {
                            const token = this.tokens[i];
                            const r = TSParametersConversions[token.type](m[i + 1]);
                            if (!(0, commons_1.$ok)(r)) {
                                throw new tserrors_1.TSHttpError(`Bad parameter '${token.name}' of type '${token.type}' in url path '${path}'`, tsrequest_1.Resp.BadRequest, {
                                    token: token.name,
                                    type: token.type,
                                    path: path
                                });
                            }
                            ret[token.name] = r;
                        }
                    }
                    return ret;
                }
            }
        }
        return null;
    }
}
exports.TSParametricEndPoints = TSParametricEndPoints;
TSParametricEndPoints.__verbs = Object.values(tsrequest_1.Verb);
function _calculateQuery(req, qdef) {
    let query = {};
    req.url.searchParams.forEach((v, k) => {
        k = (0, commons_1.$ftrim)(k).toLowerCase();
        if (!(0, commons_1.$length)(k)) {
            throw new tserrors_1.TSHttpError(`Unamed query key for endpoint ${req.method} '${req.url.pathname}'`, tsrequest_1.Resp.BadRequest, {
                method: req.method,
                path: req.url.pathname
            });
        }
        const queryItem = qdef[k];
        if (!(0, commons_1.$ok)(queryItem)) {
            throw new tserrors_1.TSHttpError(`Unexpected query key '${k}' for endpoint ${req.method} '${req.url.pathname}'`, tsrequest_1.Resp.BadRequest, {
                method: req.method,
                path: req.url.pathname,
                queryKey: k
            });
        }
        if ((0, commons_1.$length)(v)) {
            const r = TSParametersConversions[queryItem.type](v);
            if (!(0, commons_1.$ok)(r)) {
                throw new tserrors_1.TSHttpError(`Bad query value type for key '${k}' for endpoint ${req.method} '${req.url.pathname}'`, tsrequest_1.Resp.BadRequest, {
                    method: req.method,
                    path: req.url.pathname,
                    queryKey: k,
                    queryValue: v
                });
            }
            query[k] = r;
        }
    });
    const possibleItems = (0, commons_1.$keys)(qdef);
    possibleItems.forEach(qn => {
        if (qdef[qn].mandatory && !(0, commons_1.$ok)(query[qn])) {
            throw new tserrors_1.TSHttpError(`Mandatory query value for key '${qn}' is missing for endpoint ${req.method} '${req.url.pathname}'`, tsrequest_1.Resp.BadRequest, {
                method: req.method,
                path: req.url.pathname,
                queryKey: qn
            });
        }
    });
    req.query = query;
}
//# sourceMappingURL=tsservercomp.js.map